@@ -2,7 +2,7 @@
 /***************************************************************************
  * Copyright (C) 2005 David Zeuthen, <davidz@redhat.com>
  * Copyright (C) 2005 - 2008 Dan Williams, <dcbw@redhat.com>
- * Copyright (C) 2005 - 2011 Red Hat, Inc.
+ * Copyright (C) 2005 - 2021 Red Hat, Inc.
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -229,6 +229,11 @@ import (NMVpnEditorPlugin *iface, const char *path, GError **error)
 	if (buf)
 		nm_setting_vpn_add_data_item (s_vpn, NM_OPENCONNECT_KEY_PROXY, buf);
 
+	/* UserAgent */
+	buf = g_key_file_get_string (keyfile, "openconnect", "UserAgent", NULL);
+	if (buf)
+		nm_setting_vpn_add_data_item (s_vpn, NM_OPENCONNECT_KEY_USERAGENT, buf);
+
 	/* Cisco Secure Desktop */
 	bval = g_key_file_get_boolean (keyfile, "openconnect", "CSDEnable", NULL);
 	if (bval)
@@ -244,12 +249,23 @@ import (NMVpnEditorPlugin *iface, const char *path, GError **error)
 	if (buf)
 		nm_setting_vpn_add_data_item (s_vpn, NM_OPENCONNECT_KEY_REPORTED_OS, buf);
 
-	/* User Certificate */
+	/* Machine Certificate */
+	buf = g_key_file_get_string(keyfile, "openconnect", "MachineCertificate",
+								NULL);
+	if (buf && strcmp(buf, "(null)"))
+		nm_setting_vpn_add_data_item(s_vpn, NM_OPENCONNECT_KEY_MCACERT, buf);
+
+	/* MCA Private Key */
+	buf = g_key_file_get_string(keyfile, "openconnect", "MachinePrivateKey", NULL);
+	if (buf && strcmp(buf, "(null)"))
+		nm_setting_vpn_add_data_item(s_vpn, NM_OPENCONNECT_KEY_MCAKEY, buf);
+
+    /* User Certificate */
 	buf = g_key_file_get_string (keyfile, "openconnect", "UserCertificate", NULL);
 	if (buf && strcmp(buf, "(null)"))
 		nm_setting_vpn_add_data_item (s_vpn, NM_OPENCONNECT_KEY_USERCERT, buf);
 
-	/* Private Key */
+	/* User Private Key */
 	buf = g_key_file_get_string (keyfile, "openconnect", "PrivateKey", NULL);
 	if (buf && strcmp(buf, "(null)"))
 		nm_setting_vpn_add_data_item (s_vpn, NM_OPENCONNECT_KEY_PRIVKEY, buf);
@@ -258,12 +274,17 @@ import (NMVpnEditorPlugin *iface, const char *path, GError **error)
 	bval = g_key_file_get_boolean (keyfile, "openconnect", "FSID", NULL);
 	if (bval)
 		nm_setting_vpn_add_data_item (s_vpn, NM_OPENCONNECT_KEY_PEM_PASSPHRASE_FSID, "yes");
-	
+
 	/* Prevent invalid cert */
 	bval = g_key_file_get_boolean (keyfile, "openconnect", "PreventInvalidCert", NULL);
-	if (true)
+	if (bval)
 		nm_setting_vpn_add_data_item (s_vpn, NM_OPENCONNECT_KEY_PREVENT_INVALID_CERT, "yes");
 
+	/* Disable UDP (DTLS and ESP) and use only TLS tunnel */
+	bval = g_key_file_get_boolean (keyfile, "openconnect", "DisableUDP", NULL);
+	if (bval)
+		nm_setting_vpn_add_data_item (s_vpn, NM_OPENCONNECT_KEY_DISABLE_UDP, "yes");
+
 	/* Soft token mode */
 	buf = g_key_file_get_string (keyfile, "openconnect", "StokenSource", NULL);
 	if (buf)
@@ -290,13 +311,17 @@ export (NMVpnEditorPlugin *iface,
 	const char *cacert = NULL;
 	const char *protocol = NULL;
 	const char *proxy = NULL;
+	const char *useragent = NULL;
 	gboolean csd_enable = FALSE;
 	const char *csd_wrapper = NULL;
 	const char *reported_os = NULL;
+	const char *mcacert = NULL;
+	const char *mcaprivkey = NULL;
 	const char *usercert = NULL;
 	const char *privkey = NULL;
 	gboolean pem_passphrase_fsid = FALSE;
 	gboolean prevent_invalid_cert = FALSE;
+	gboolean disable_udp = FALSE;
 	const char *token_mode = NULL;
 	const char *token_secret = NULL;
 	gboolean success = FALSE;
@@ -338,6 +363,10 @@ export (NMVpnEditorPlugin *iface,
 	if (value && strlen (value))
 		proxy = value;
 
+	value = nm_setting_vpn_get_data_item(s_vpn, NM_OPENCONNECT_KEY_USERAGENT);
+	if (value && strlen(value))
+		useragent = value;
+
 	value = nm_setting_vpn_get_data_item (s_vpn, NM_OPENCONNECT_KEY_CSD_ENABLE);
 	if (value && !strcmp (value, "yes"))
 		csd_enable = TRUE;
@@ -350,6 +379,14 @@ export (NMVpnEditorPlugin *iface,
 	if (value && strlen (value))
 		reported_os = value;
 
+	value = nm_setting_vpn_get_data_item (s_vpn, NM_OPENCONNECT_KEY_MCACERT);
+	if (value && strlen (value))
+		mcacert = value;
+
+	value = nm_setting_vpn_get_data_item (s_vpn, NM_OPENCONNECT_KEY_MCAKEY);
+	if (value && strlen (value))
+		mcaprivkey = value;
+
 	value = nm_setting_vpn_get_data_item (s_vpn, NM_OPENCONNECT_KEY_USERCERT);
 	if (value && strlen (value))
 		usercert = value;
@@ -361,11 +398,15 @@ export (NMVpnEditorPlugin *iface,
 	value = nm_setting_vpn_get_data_item (s_vpn, NM_OPENCONNECT_KEY_PEM_PASSPHRASE_FSID);
 	if (value && !strcmp (value, "yes"))
 		pem_passphrase_fsid = TRUE;
-	
+
 	value = nm_setting_vpn_get_data_item (s_vpn, NM_OPENCONNECT_KEY_PREVENT_INVALID_CERT);
 	if (value && !strcmp (value, "yes"))
 		prevent_invalid_cert = TRUE;
 
+	value = nm_setting_vpn_get_data_item (s_vpn, NM_OPENCONNECT_KEY_DISABLE_UDP);
+	if (value && !strcmp (value, "yes"))
+		disable_udp = TRUE;
+
 	value = nm_setting_vpn_get_data_item (s_vpn, NM_OPENCONNECT_KEY_TOKEN_MODE);
 	if (value && strlen (value))
 		token_mode = value;
@@ -386,13 +427,17 @@ export (NMVpnEditorPlugin *iface,
 		 "CACert=%s\n"
 		 "Protocol=%s\n"
 		 "Proxy=%s\n"
+		 "UserAgent=%s\n"
 		 "CSDEnable=%s\n"
 		 "CSDWrapper=%s\n"
 		 "ReportedOS=%s\n"
+		 "MachineCertificate=%s\n"
+		 "MachinePrivateKey=%s\n"
 		 "UserCertificate=%s\n"
 		 "PrivateKey=%s\n"
 		 "FSID=%s\n"
 		 "PreventInvalidCert=%s\n"
+		 "DisableUDP=%s\n"
 		 "StokenSource=%s\n"
 		 "StokenString=%s\n",
 		 /* Description */           nm_setting_connection_get_id (s_con),
@@ -400,13 +445,17 @@ export (NMVpnEditorPlugin *iface,
 		 /* CA Certificate */        cacert ? cacert : "",
 		 /* Protocol */              protocol ? protocol : "anyconnect",
 		 /* Proxy */                 proxy ? proxy : "",
+		 /* User Agent */            useragent ? useragent : "",
 		 /* Cisco Secure Desktop */  csd_enable ? "1" : "0",
 		 /* CSD Wrapper Script */    csd_wrapper ? csd_wrapper : "",
 		 /* Reported OS */           reported_os ? reported_os : "",
+		 /* MCA Certificate */       mcacert ? mcacert : "",
+		 /* MCA Private Key */       mcaprivkey ? mcaprivkey : "",
 		 /* User Certificate */      usercert ? usercert : "",
-		 /* Private Key */           privkey ? privkey : "",
+		 /* User Private Key */      privkey ? privkey : "",
 		 /* FSID */                  pem_passphrase_fsid ? "1" : "0",
 		 /* Prevent invalid cert */  prevent_invalid_cert ? "1" : "0",
+		 /* Disable UDP */           disable_udp ? "1" : "0",
 		 /* Soft token mode */       token_mode ? token_mode : "",
 		 /* Soft token secret */     token_secret ? token_secret : "");
 
@@ -515,23 +564,35 @@ _call_editor_factory (gpointer factory,
 static NMVpnEditor *
 get_editor (NMVpnEditorPlugin *iface, NMConnection *connection, GError **error)
 {
+	gpointer gtk3_only_symbol;
+	GModule *self_module;
+	const char *editor;
+
 	g_return_val_if_fail (OPENCONNECT_IS_EDITOR_PLUGIN (iface), NULL);
 	g_return_val_if_fail (NM_IS_CONNECTION (connection), NULL);
 	g_return_val_if_fail (!error || !*error, NULL);
 
-	{
-#ifdef NM_VPN_OLD
-		return nm_vpn_editor_new (connection, error);
-#else
-		return nm_vpn_plugin_utils_load_editor (NM_PLUGIN_DIR"/libnm-vpn-plugin-openconnect-editor.so",
-		                                        "nm_vpn_editor_factory_openconnect",
-		                                        _call_editor_factory,
-		                                        iface,
-		                                        connection,
-		                                        NULL,
-		                                        error);
-#endif
+	self_module = g_module_open (NULL, 0);
+	g_module_symbol (self_module, "gtk_container_add", &gtk3_only_symbol);
+	g_module_close (self_module);
+
+	if (gtk3_only_symbol) {
+		editor = "libnm-vpn-plugin-openconnect-editor.so";
+	} else {
+		editor = "libnm-gtk4-vpn-plugin-openconnect-editor.so";
 	}
+
+#ifdef NM_VPN_OLD
+	return nm_vpn_editor_new (connection, error);
+#else
+	return nm_vpn_plugin_utils_load_editor (editor,
+	                                        "nm_vpn_editor_factory_openconnect",
+	                                        _call_editor_factory,
+	                                        iface,
+	                                        connection,
+	                                        NULL,
+	                                        error);
+#endif
 }
 
 static void
@@ -621,4 +682,3 @@ nm_vpn_editor_plugin_factory (GError **error)
 
 	return g_object_new (OPENCONNECT_TYPE_EDITOR_PLUGIN, NULL);
 }
-
