@@ -47,6 +47,23 @@
 #define openconnect_has_yubioath_support() 0
 #endif
 
+#if !OPENCONNECT_CHECK_VER(5,5)
+#define OC_PROTO_PROXY				(1<<0)
+#define OC_PROTO_CSD				(1<<1)
+#define OC_PROTO_AUTH_CERT			(1<<2)
+#define OC_PROTO_AUTH_OTP			(1<<3)
+#define OC_PROTO_AUTH_STOKEN		(1<<4)
+#endif
+
+#if !OPENCONNECT_CHECK_VER(5,7)
+#define OC_PROTO_PERIODIC_TROJAN	(1<<5)
+#define OC_PROTO_HIDDEN				(1<<6)
+#endif
+
+#if !OPENCONNECT_CHECK_VER(5,8)
+#define OC_PROTO_AUTH_MCA			(1<<7)
+#endif
+
 #include "auth-helpers.h"
 
 #if !GTK_CHECK_VERSION(4,0,0)
@@ -250,6 +267,7 @@ init_protocol_combo_options (GtkComboBox *protocol_combo)
 		gtk_list_store_set(protocol_combo_list, &iter,
 						   0, protos[i].pretty_name,
 						   1, protos[i].name,
+						   2, protos[i].flags,
 						   -1);
 	}
 	openconnect_free_supported_protocols(protos);
@@ -258,12 +276,14 @@ init_protocol_combo_options (GtkComboBox *protocol_combo)
 	gtk_list_store_set(protocol_combo_list, &iter,
 					   0, _("Cisco AnyConnect"),
 					   1, "anyconnect",
+					   2, OC_PROTO_PROXY | OC_PROTO_CSD | OC_PROTO_AUTH_CERT | OC_PROTO_AUTH_OTP | OC_PROTO_AUTH_STOKEN,
 					   -1);
 #  if OPENCONNECT_CHECK_VER(5,1)
 	gtk_list_store_append(protocol_combo_list, &iter);
 	gtk_list_store_set(protocol_combo_list, &iter,
 					   0, _("Juniper/Pulse Network Connect"),
 					   1, "nc",
+					   2, OC_PROTO_PROXY | OC_PROTO_CSD | OC_PROTO_AUTH_CERT | OC_PROTO_AUTH_OTP | OC_PROTO_AUTH_STOKEN,
 					   -1);
 #  endif
 #endif
@@ -344,6 +364,16 @@ init_editor_plugin (OpenconnectEditor *self, NMConnection *connection, GError **
 	}
 	g_signal_connect (G_OBJECT (widget), "changed", G_CALLBACK (stuff_changed_cb), self);
 
+	widget = GTK_WIDGET (gtk_builder_get_object (priv->builder, "user_agent_entry"));
+	g_return_val_if_fail (widget, FALSE);
+
+	if (s_vpn) {
+		value = nm_setting_vpn_get_data_item (s_vpn, NM_OPENCONNECT_KEY_USERAGENT);
+		if (value)
+			gtk_editable_set_text (GTK_EDITABLE (widget), value);
+	}
+	g_signal_connect (G_OBJECT (widget), "changed", G_CALLBACK (stuff_changed_cb), self);
+
 	widget = GTK_WIDGET (gtk_builder_get_object (priv->builder, "fsid_button"));
 	g_return_val_if_fail (widget, FALSE);
 
@@ -364,6 +394,16 @@ init_editor_plugin (OpenconnectEditor *self, NMConnection *connection, GError **
 	}
 	g_signal_connect (G_OBJECT (widget), "toggled", G_CALLBACK (stuff_changed_cb), self);
 
+	widget = GTK_WIDGET (gtk_builder_get_object (priv->builder, "disable_udp_button"));
+	g_return_val_if_fail (widget, FALSE);
+
+	if (s_vpn) {
+		value = nm_setting_vpn_get_data_item (s_vpn, NM_OPENCONNECT_KEY_DISABLE_UDP);
+		if (value && !strcmp(value, "yes"))
+			gtk_check_button_set_active(GTK_CHECK_BUTTON (widget), TRUE);
+	}
+	g_signal_connect (G_OBJECT (widget), "toggled", G_CALLBACK (stuff_changed_cb), self);
+
 	widget = GTK_WIDGET (gtk_builder_get_object (priv->builder, "csd_button"));
 	g_return_val_if_fail (widget, FALSE);
 
@@ -445,9 +485,18 @@ update_connection (NMVpnEditor *iface,
 	widget = GTK_WIDGET (gtk_builder_get_object (priv->builder, "protocol_combo"));
 	model = gtk_combo_box_get_model (GTK_COMBO_BOX (widget));
 	if (gtk_combo_box_get_active_iter (GTK_COMBO_BOX (widget), &iter)) {
-		gtk_tree_model_get (model, &iter, 1, &str, -1);
+		gint flags = 0;
+		gtk_tree_model_get (model, &iter, 1, &str, 2, &flags, -1);
 		nm_setting_vpn_add_data_item (s_vpn, NM_OPENCONNECT_KEY_PROTOCOL, str);
 		g_free(str);
+
+		widget = GTK_WIDGET (gtk_builder_get_object (priv->builder, "mca_cert_chooser"));
+		if (widget) {
+			if (flags & OC_PROTO_AUTH_MCA)
+				gtk_widget_show(widget);
+			else
+				gtk_widget_hide(widget);
+		}
 	}
 
 	widget = GTK_WIDGET (gtk_builder_get_object (priv->builder, "gateway_entry"));
@@ -460,6 +509,11 @@ update_connection (NMVpnEditor *iface,
 	if (str && strlen (str))
 		nm_setting_vpn_add_data_item (s_vpn, NM_OPENCONNECT_KEY_PROXY, str);
 
+	widget = GTK_WIDGET (gtk_builder_get_object (priv->builder, "user_agent_entry"));
+	str = (char *) gtk_editable_get_text (GTK_EDITABLE (widget));
+	if (str && strlen (str))
+		nm_setting_vpn_add_data_item (s_vpn, NM_OPENCONNECT_KEY_USERAGENT, str);
+
 	widget = GTK_WIDGET (gtk_builder_get_object (priv->builder, "fsid_button"));
 	str = gtk_check_button_get_active (GTK_CHECK_BUTTON (widget))?"yes":"no";
 	nm_setting_vpn_add_data_item (s_vpn, NM_OPENCONNECT_KEY_PEM_PASSPHRASE_FSID, str);
@@ -468,6 +522,10 @@ update_connection (NMVpnEditor *iface,
 	str = gtk_check_button_get_active (GTK_CHECK_BUTTON (widget))?"yes":"no";
 	nm_setting_vpn_add_data_item (s_vpn, NM_OPENCONNECT_KEY_PREVENT_INVALID_CERT, str);
 
+	widget = GTK_WIDGET (gtk_builder_get_object (priv->builder, "disable_udp_button"));
+	str = gtk_check_button_get_active(GTK_CHECK_BUTTON (widget))?"yes":"no";
+	nm_setting_vpn_add_data_item (s_vpn, NM_OPENCONNECT_KEY_DISABLE_UDP, str);
+
 	widget = GTK_WIDGET (gtk_builder_get_object (priv->builder, "csd_button"));
 	str = gtk_check_button_get_active (GTK_CHECK_BUTTON (widget))?"yes":"no";
 	nm_setting_vpn_add_data_item (s_vpn, NM_OPENCONNECT_KEY_CSD_ENABLE, str);
