@@ -88,6 +88,7 @@ static const ValidProperty valid_properties[] = {
 	{ NM_OPENCONNECT_KEY_MTU,         G_TYPE_STRING, 0, 0 },
 	{ NM_OPENCONNECT_KEY_PEM_PASSPHRASE_FSID, G_TYPE_BOOLEAN, 0, 0 },
 	{ NM_OPENCONNECT_KEY_PREVENT_INVALID_CERT, G_TYPE_BOOLEAN, 0, 0 },
+	{ NM_OPENCONNECT_KEY_DISABLE_UDP, G_TYPE_BOOLEAN, 0, 0 },
 	{ NM_OPENCONNECT_KEY_PROTOCOL,    G_TYPE_STRING, 0, 0 },
 	{ NM_OPENCONNECT_KEY_PROXY,       G_TYPE_STRING, 0, 0 },
 	{ NM_OPENCONNECT_KEY_CSD_ENABLE,  G_TYPE_BOOLEAN, 0, 0 },
@@ -95,6 +96,9 @@ static const ValidProperty valid_properties[] = {
 	{ NM_OPENCONNECT_KEY_TOKEN_MODE,  G_TYPE_STRING, 0, 0 },
 	{ NM_OPENCONNECT_KEY_TOKEN_SECRET, G_TYPE_STRING, 0, 0 },
 	{ NM_OPENCONNECT_KEY_REPORTED_OS, G_TYPE_STRING, 0, 0 },
+	{ NM_OPENCONNECT_KEY_MCACERT,     G_TYPE_STRING, 0, 0 },
+	{ NM_OPENCONNECT_KEY_MCAKEY,      G_TYPE_STRING, 0, 0 },
+	{ NM_OPENCONNECT_KEY_MCA_PASS,    G_TYPE_STRING, 0, 0 },
 	{ NULL,                           G_TYPE_NONE, 0, 0 }
 };
 
@@ -102,6 +106,7 @@ static const ValidProperty valid_secrets[] = {
 	{ NM_OPENCONNECT_KEY_COOKIE,  G_TYPE_STRING, 0, 0 },
 	{ NM_OPENCONNECT_KEY_GATEWAY, G_TYPE_STRING, 0, 0 },
 	{ NM_OPENCONNECT_KEY_GWCERT,  G_TYPE_STRING, 0, 0 },
+	{ NM_OPENCONNECT_KEY_RESOLVE, G_TYPE_STRING, 0, 0 },
 	{ NULL,                       G_TYPE_NONE, 0, 0 }
 };
 
@@ -249,7 +254,7 @@ nm_openconnect_secrets_validate (NMSettingVpn *s_vpn, GError **error)
 }
 
 static char *
-create_persistent_tundev(void)
+create_persistent_tundev(const char *suggested_name, GError **error)
 {
 	struct passwd *pw;
 	struct ifreq ifr;
@@ -257,8 +262,11 @@ create_persistent_tundev(void)
 	int i;
 
 	pw = getpwnam(NM_OPENCONNECT_USER);
-	if (!pw)
+	if (!pw) {
+		/* TODO: Is this an error case or a normal possibility
+		 * where we just don't set a persistent tundev? */
 		return NULL;
+	}
 
 	gl.tun_owner = pw->pw_uid;
 	gl.tun_group = pw->pw_gid;
@@ -266,33 +274,67 @@ create_persistent_tundev(void)
 	fd = open("/dev/net/tun", O_RDWR);
 	if (fd < 0) {
 		perror("open /dev/net/tun");
-		exit(EXIT_FAILURE);
+		g_set_error (error,
+		             NM_VPN_PLUGIN_ERROR,
+		             NM_VPN_PLUGIN_ERROR_LAUNCH_FAILED,
+		             "%s",
+		             _("Could not open /dev/net/tun"));
+		return NULL;
 	}
 
 	memset(&ifr, 0, sizeof(ifr));
 	ifr.ifr_flags = IFF_TUN | IFF_NO_PI;
 
-	for (i = 0; i < 256; i++) {
-		sprintf(ifr.ifr_name, "vpn%d", i);
+	if (suggested_name) {
+		/* Not NUL-terminated if strlen(suggested_name) >= IFNAMSIZE  */
+		strncpy(ifr.ifr_name, suggested_name, IFNAMSIZ);
+		if (ioctl(fd, TUNSETIFF, (void *)&ifr)) {
+			g_set_error (error,
+			             NM_VPN_PLUGIN_ERROR,
+			             NM_VPN_PLUGIN_ERROR_LAUNCH_FAILED,
+			             _("Could not create tunnel interface named '%.*s'"),
+			             IFNAMSIZ, ifr.ifr_name);
+			return NULL;
+		}
+	} else {
+		for (i = 0; i < 256; i++) {
+			sprintf(ifr.ifr_name, "vpn%d", i);
 
-		if (!ioctl(fd, TUNSETIFF, (void *)&ifr))
-			break;
+			if (!ioctl(fd, TUNSETIFF, (void *)&ifr))
+				break;
+		}
+		if (i == 256) {
+			g_set_error (error,
+			             NM_VPN_PLUGIN_ERROR,
+			             NM_VPN_PLUGIN_ERROR_LAUNCH_FAILED,
+			             "%s",
+			             _("Could not create tunnel interface named 'vpnX', for X in 0..255"));
+			return NULL;
+		}
 	}
-	if (i == 256)
-		exit(EXIT_FAILURE);
 
 	if (ioctl(fd, TUNSETOWNER, gl.tun_owner) < 0) {
 		perror("TUNSETOWNER");
-		exit(EXIT_FAILURE);
+		g_set_error (error,
+		             NM_VPN_PLUGIN_ERROR,
+		             NM_VPN_PLUGIN_ERROR_LAUNCH_FAILED,
+		             _("Could not set tunnel interface owner to '%d'"),
+		             gl.tun_owner);
+		return NULL;
 	}
 
 	if (ioctl(fd, TUNSETPERSIST, 1)) {
 		perror("TUNSETPERSIST");
-		exit(EXIT_FAILURE);
+		g_set_error (error,
+		             NM_VPN_PLUGIN_ERROR,
+		             NM_VPN_PLUGIN_ERROR_LAUNCH_FAILED,
+		             "%s",
+		             _("Could not set tunnel interface persistence"));
+		return NULL;
 	}
 	close(fd);
 	_LOGW ("Created tundev %s\n", ifr.ifr_name);
-	return g_strdup(ifr.ifr_name);
+	return g_strndup(ifr.ifr_name, IFNAMSIZ);
 }
 
 static void
@@ -384,6 +426,7 @@ openconnect_watch_cb (GPid pid, gint status, gpointer user_data)
 
 static gint
 nm_openconnect_start_openconnect_binary (NMOpenconnectPlugin *plugin,
+                                         NMSettingConnection *s_con,
                                          NMSettingVpn *s_vpn,
                                          GError **error)
 {
@@ -395,8 +438,10 @@ nm_openconnect_start_openconnect_binary (NMOpenconnectPlugin *plugin,
 	gint stdin_fd;
 	char csd_user_arg[60];
 	const char *props_vpn_gw, *props_cookie, *props_cacert, *props_mtu, *props_gwcert, *props_proxy;
-	const char *props_csd_enable, *props_csd_wrapper;
+	const char *props_csd_enable, *props_csd_wrapper, *props_resolve;
+	const char *props_disable_udp;
 	const char *protocol;
+	const char *props_tun_name;
 
 	/* Find openconnect */
 	openconnect_binary = openconnect_binary_paths;
@@ -437,6 +482,7 @@ nm_openconnect_start_openconnect_binary (NMOpenconnectPlugin *plugin,
 		return -1;
 	}
 	props_gwcert = nm_setting_vpn_get_secret (s_vpn, NM_OPENCONNECT_KEY_GWCERT);
+	props_resolve = nm_setting_vpn_get_secret (s_vpn, NM_OPENCONNECT_KEY_RESOLVE);
 
 	props_cacert = nm_setting_vpn_get_data_item (s_vpn, NM_OPENCONNECT_KEY_CACERT);
 	props_mtu = nm_setting_vpn_get_data_item (s_vpn, NM_OPENCONNECT_KEY_MTU);
@@ -481,10 +527,22 @@ nm_openconnect_start_openconnect_binary (NMOpenconnectPlugin *plugin,
 	g_ptr_array_add (openconnect_argv, (gpointer) "--script");
 	g_ptr_array_add (openconnect_argv, (gpointer) NM_OPENCONNECT_HELPER_PATH);
 
-	priv->tun_name = create_persistent_tundev ();
-	if (priv->tun_name) {
+	props_tun_name = nm_setting_connection_get_interface_name(s_con);
+	priv->tun_name = create_persistent_tundev (props_tun_name, error);
+	if (!priv->tun_name && *error)
+		return -1;
+	else if (priv->tun_name) {
 		g_ptr_array_add (openconnect_argv, (gpointer) "--interface");
 		g_ptr_array_add (openconnect_argv, (gpointer) priv->tun_name);
+	} else {
+		/* TODO: Is this a valid case? See create_persistent_tundev. */
+	}
+
+	props_disable_udp = nm_setting_vpn_get_data_item (s_vpn, NM_OPENCONNECT_KEY_DISABLE_UDP);
+	if (props_disable_udp && !strcmp (props_disable_udp, "yes")) {
+		/* May be renamed to --no-udp in the future, but with --no-dtls kept for backwards-
+		 * compatibility (see https://gitlab.com/openconnect/openconnect/-/merge_requests/151) */
+		g_ptr_array_add (openconnect_argv, (gpointer) "--no-dtls");
 	}
 
 	props_csd_enable = nm_setting_vpn_get_data_item (s_vpn, NM_OPENCONNECT_KEY_CSD_ENABLE);
@@ -492,7 +550,7 @@ nm_openconnect_start_openconnect_binary (NMOpenconnectPlugin *plugin,
 	if (props_csd_enable && !strcmp (props_csd_enable, "yes") && props_csd_wrapper) {
 		/* TODO: disable passing the script to openconnect.
 		 *
-		 * As we have priv->tun_name, openconnect will run as an unpriviledged user NM_OPENCONNECT_USER.
+		 * As we have priv->tun_name, openconnect will run as an unprivileged user NM_OPENCONNECT_USER.
 		 * However, it is still not safe to run untrusted scripts provided by the user.
 		 *
 		 * This needs a different solution, for now, just log a warning. */
@@ -512,6 +570,11 @@ nm_openconnect_start_openconnect_binary (NMOpenconnectPlugin *plugin,
 
 	g_ptr_array_add (openconnect_argv, (gpointer) props_vpn_gw);
 
+	if (props_resolve && strlen(props_resolve)) {
+		g_ptr_array_add (openconnect_argv, (gpointer) "--resolve");
+		g_ptr_array_add (openconnect_argv, (gpointer) props_resolve);
+	}
+
 	if (gl.log_level >= LOG_INFO) {
 		g_ptr_array_add (openconnect_argv, (gpointer) "--verbose");
 		if (gl.log_level >= LOG_DEBUG)
@@ -553,9 +616,14 @@ real_connect (NMVpnServicePlugin   *plugin,
               NMConnection  *connection,
               GError       **error)
 {
+	NMSettingConnection *s_con;
 	NMSettingVpn *s_vpn;
 	gint openconnect_fd = -1;
 
+	s_con = nm_connection_get_setting_connection(connection);
+	g_assert (s_con);
+	if (!s_con)
+		goto out;
 	s_vpn = nm_connection_get_setting_vpn (connection);
 	g_assert (s_vpn);
 	if (!nm_openconnect_properties_validate (s_vpn, error))
@@ -566,7 +634,7 @@ real_connect (NMVpnServicePlugin   *plugin,
 	if (_LOGD_enabled ())
 		nm_connection_dump (connection);
 
-	openconnect_fd = nm_openconnect_start_openconnect_binary (NM_OPENCONNECT_PLUGIN (plugin), s_vpn, error);
+	openconnect_fd = nm_openconnect_start_openconnect_binary (NM_OPENCONNECT_PLUGIN (plugin), s_con, s_vpn, error);
 	if (!openconnect_fd)
 		return TRUE;
 
@@ -596,7 +664,7 @@ real_need_secrets (NMVpnServicePlugin *plugin,
 	}
 
 	/* We just need the WebVPN cookie, and the final IP address of the gateway
-	   (after HTTP redirects, which do happen). All the certificate/SecurID 
+	   (after HTTP redirects, which do happen). All the certificate/SecurID
 	   nonsense can be handled for us, in the user's context, by auth-dialog */
 	if (!nm_setting_vpn_get_secret (s_vpn, NM_OPENCONNECT_KEY_GATEWAY)) {
 		*setting_name = NM_SETTING_VPN_SETTING_NAME;
